package uk.ac.leeds.ccg.andyt.vector.geometry;

import java.math.BigDecimal;

public class Envelope2D
        extends AbstractGeometry2D {

    BigDecimal _xmin;
    BigDecimal _xmax;
    BigDecimal _ymin;
    BigDecimal _ymax;

    public Envelope2D(
            Envelope2D aEnvelope2D) {
        _ymin = new BigDecimal(aEnvelope2D._ymin.toString());
        _ymax = new BigDecimal(aEnvelope2D._ymax.toString());
        _xmin = new BigDecimal(aEnvelope2D._xmin.toString());
        _xmax = new BigDecimal(aEnvelope2D._xmax.toString());
        applyDecimalPlacePrecision();
    }

    public Envelope2D(
            AbstractGeometry2D aAbstractGeometry2D) {
        Envelope2D aEnvelope2D = aAbstractGeometry2D.getEnvelope2D();
        _xmin = aEnvelope2D._xmin;
        _xmax = aEnvelope2D._xmin;
        _ymin = aEnvelope2D._ymin;
        _ymax = aEnvelope2D._ymax;
        applyDecimalPlacePrecision();
    }

    public Envelope2D(
            Point2D aPoint,
            Point2D bPoint) {
        if (aPoint._y.compareTo(bPoint._y) == 1) {
            _ymin = bPoint._y;
            _ymax = aPoint._y;
        } else {
            _ymin = aPoint._y;
            _ymax = bPoint._y;
        }
        if (aPoint._x.compareTo(bPoint._x) == 1) {
            _xmin = bPoint._x;
            _xmax = aPoint._x;
        } else {
            _xmin = aPoint._x;
            _xmax = bPoint._x;
        }
        applyDecimalPlacePrecision();
    }

    public Envelope2D(
            BigDecimal x,
            BigDecimal y) {
        _ymin = new BigDecimal(y.toString());
        _ymax = new BigDecimal(y.toString());
        _xmin = new BigDecimal(x.toString());
        _xmax = new BigDecimal(x.toString());
        applyDecimalPlacePrecision();
    }

    public Envelope2D(
            AbstractGeometry2D[] aAbstractGeometry2DArray) {
        Envelope2D aEnvelope2D = aAbstractGeometry2DArray[0].getEnvelope2D();
        _xmin = aEnvelope2D._xmin;
        _xmax = aEnvelope2D._xmin;
        _ymin = aEnvelope2D._ymin;
        _ymax = aEnvelope2D._ymax;
        for (int i = 1; i < aAbstractGeometry2DArray.length; i++) {
            aEnvelope2D = aAbstractGeometry2DArray[i].getEnvelope2D();
            if (aEnvelope2D._xmin.compareTo(_xmin) == -1) {
                _xmin = aEnvelope2D._xmin;
            }
            if (aEnvelope2D._xmax.compareTo(_xmax) == 1) {
                _xmax = aEnvelope2D._xmax;
            }
            if (aEnvelope2D._ymin.compareTo(_ymin) == -1) {
                _ymin = aEnvelope2D._ymin;
            }
            if (aEnvelope2D._ymax.compareTo(_ymax) == 1) {
                _ymax = aEnvelope2D._ymax;
            }
        }
        applyDecimalPlacePrecision();
    }

    @Override
    public String toString() {
        return new String(
                "Envelope2D(" +
                super.toString() +
                "xmin(" + _xmin.toString() + ")," +
                "xmax(" + _xmax.toString() + ")," +
                "ymin(" + _ymin.toString() + ")," +
                "ymax(" + _xmax.toString() + "))");
    }

    /**
     * If aEnvelope2D touches, or overlaps then it intersects.
     */
    public boolean intersects(Envelope2D aEnvelope2D) {
        return getIntersects(
                aEnvelope2D._xmin,
                aEnvelope2D._ymin) ||
                getIntersects(
                aEnvelope2D._xmin,
                aEnvelope2D._ymax) ||
                getIntersects(
                aEnvelope2D._xmax,
                aEnvelope2D._ymin) ||
                getIntersects(
                aEnvelope2D._xmax,
                aEnvelope2D._ymax);
//        Point2D aPoint2D = new Point2D(
//                aEnvelope2D._xmin,
//                aEnvelope2D._ymin);
//        Point2D bPoint2D = new Point2D(
//                aEnvelope2D._xmin,
//                aEnvelope2D._ymax);
//        Point2D cPoint2D = new Point2D(
//                aEnvelope2D._xmax,
//                aEnvelope2D._ymin);
//        Point2D dPoint2D = new Point2D(
//                aEnvelope2D._xmax,
//                aEnvelope2D._ymax);       
//        return intersects(aPoint2D) ||
//                intersects(bPoint2D) ||
//                intersects(cPoint2D) ||
//                intersects(dPoint2D);
    }

    /**
     *
     * @param a_LineSegment2D
     * @return 0 if no, 1 if yes, 2 if maybe.
     */
    public int getIntersects(LineSegment2D a_LineSegment2D) {
        Envelope2D a_Envelope2D = a_LineSegment2D.getEnvelope2D();
        if (a_Envelope2D.intersects(getEnvelope2D())){
            if (getIntersects(a_LineSegment2D._Start_Point2D)){
                return 1;
            } else {
                if(getIntersects(a_LineSegment2D._End_Point2D)){
                    return 1;
                }
                return 2;
            }
        } else { 
                 return 0;
        }
    }

    public boolean getIntersects(Point2D aPoint) {
       return aPoint._x.compareTo(_xmin) != -1 &&
                aPoint._x.compareTo(_xmax) != 1 &&
                aPoint._y.compareTo(_ymin) != -1 &&
                aPoint._y.compareTo(_ymax) != 1;
    }

    public boolean getIntersects(
            BigDecimal x,
            BigDecimal y) {
        return x.compareTo(_xmin) != -1 &&
                x.compareTo(_xmax) != 1 &&
                y.compareTo(_ymin) != -1 &&
                y.compareTo(_ymax) != 1;
    }

    @Override
    public Envelope2D getEnvelope2D() {
        return this;
    }

    @Override
    public void applyDecimalPlacePrecision() {
        _xmin = _xmin.setScale(
                _DecimalPlacePrecision,
                _RoundingMode);
        _xmax = _xmax.setScale(
                _DecimalPlacePrecision,
                _RoundingMode);
        _ymin = _ymin.setScale(
                _DecimalPlacePrecision,
                _RoundingMode);
        _ymax = _ymax.setScale(
                _DecimalPlacePrecision,
                _RoundingMode);
    }
}
