package uk.ac.leeds.ccg.andyt.vector.geometry;

import java.math.BigDecimal;
import java.math.MathContext;
import java.math.RoundingMode;
import java.util.HashSet;
import uk.ac.leeds.ccg.andyt.ext.math.BigDecimal0;

/**
 * A stright line connecting two points. The line has a direction as given by
 * the order of the points. Two such points are the same iff their start and
 * end points are the same.
 */
public class LineSegment2D
        extends AbstractGeometry2D
        implements Comparable {

    public Point2D _Start_Point2D;
    public Point2D _End_Point2D;

    public LineSegment2D() {
    }

    public LineSegment2D(
            Point2D _StartPoint,
            Point2D _EndPoint) {
        this._Start_Point2D = new Point2D(_StartPoint);
        this._End_Point2D = new Point2D(_EndPoint);
    }

    public LineSegment2D(
            Point2D _StartPoint,
            Point2D _EndPoint,
            int _DecimalPlacePrecision,
            RoundingMode _RoundingMode) {
        this._Start_Point2D = new Point2D(_StartPoint);
        this._End_Point2D = new Point2D(_EndPoint);
        this._DecimalPlacePrecision = _DecimalPlacePrecision;
        this._RoundingMode = _RoundingMode;
    }

    public LineSegment2D(
            LineSegment2D a_LineSegment2D) {
        this._Start_Point2D = new Point2D(a_LineSegment2D._Start_Point2D);
        this._End_Point2D = new Point2D(a_LineSegment2D._End_Point2D);
        this._DecimalPlacePrecision = a_LineSegment2D._DecimalPlacePrecision;
        this._RoundingMode = a_LineSegment2D._RoundingMode;
    }

    @Override
    public String toString() {
        return new String(
                "LineSegment2D(" + super.toString() +
                "_Start_Point2D(" + _Start_Point2D.toString() + ")" +
                "_End_Point2D(" + _End_Point2D.toString() + "))");
    }

    /**
     * Default is 1.
     * Calls compareTo on the _StartPoint.
     */
    public int compareTo(Object o) {
        if (o instanceof LineSegment2D) {
            LineSegment2D a_LineSegment2D = (LineSegment2D) o;
            int compareStart = _Start_Point2D.compareTo(a_LineSegment2D._Start_Point2D);
            if (compareStart == 0) {
                return _Start_Point2D.compareTo(a_LineSegment2D._Start_Point2D);
            } else {
                return compareStart;
            }
        }
        return 1;
    }

    public BigDecimal getLength(int numberOfDecimalPlaces) {
        BigDecimal length;
        BigDecimal xdiff = _Start_Point2D._x.subtract(_End_Point2D._x);
        BigDecimal ydiff = _Start_Point2D._y.subtract(_End_Point2D._y);
        length = BigDecimal0.sqrt0(
                xdiff.pow(2).add(ydiff.pow(2)),
                numberOfDecimalPlaces);
        return length;
    }

    @Override
    public Envelope2D getEnvelope2D() {
        return new Envelope2D(
                _Start_Point2D,
                _End_Point2D);
    }

    /**
     * Intersection done by first seeing if Envelope intersection....
     * @param aPoint
     * @return true iff a_LineSegment2D intersects this.
     */
    public boolean getIntersects(LineSegment2D a_LineSegment2D) {
        int a_Envelope2DIntersect = getEnvelope2D().getIntersects(a_LineSegment2D);
        if (a_Envelope2DIntersect == 0) {
            return false;
        } else {
            if (a_Envelope2DIntersect == 1) {
                return true;
            } else {
                int b_Envelope2DIntersect = a_LineSegment2D.getEnvelope2D().getIntersects(this);
                if (b_Envelope2DIntersect == 0) {
                    return false;
                } else {
                    if (b_Envelope2DIntersect == 1) {
                        return true;
                    } else {
                        if (getIntersection(a_LineSegment2D) != null) {
                            return true;
                        } else {
                            return false;
                        }
                    }
                }
            }
        }
    }

    /**
     * Intersection done by calculating angle or gradient of the line and
     * comparing this with that of a_Point.
     * @param aPoint
     * @return true iff a_Point2D intersects this.
     */
    public boolean getIntersects(Point2D a_Point2D) {
        if (getEnvelope2D().getIntersects(a_Point2D) == false) {
            return false;
        }
        if (_Start_Point2D._y.compareTo(_End_Point2D._y) == -1) {
            // StartPoint is Below EndPoint
            if (_Start_Point2D._x.compareTo(_End_Point2D._x) == -1) {
                // StartPoint is Left of EndPoint
                return isOnGradient(a_Point2D);
            } else {
                if (_Start_Point2D._x.compareTo(_End_Point2D._x) == 0) {
                    // StartPoint has same _x as EndPoint
                    return true;
                } else {
                    // StartPoint is Right of EndPoint
                    return isOnGradient(a_Point2D);
                }
            }
        } else {
            if (_Start_Point2D._y.compareTo(_End_Point2D._y) == 0) {
                // StartPoint has same _y as EndPoint
                return true;
            } else {
                // StartPoint is Above EndPoint
                if (_Start_Point2D._x.compareTo(_End_Point2D._x) == -1) {
                    // StartPoint is Left of EndPoint
                    return isOnGradient(a_Point2D);
                } else {
                    if (_Start_Point2D._x.compareTo(_End_Point2D._x) == 0) {
                        // StartPoint has same _x as EndPoint
                        return true;
                    } else {
                        // StartPoint is Right of EndPoint
                        return isOnGradient(a_Point2D);
                    }
                }
            }
        }
    }

    protected BigDecimal getGradient() {
        return _Start_Point2D.getGradient(_End_Point2D);
    }

    protected boolean isOnGradient(
            Point2D a_Point2D) {
        //Point2D b_Point2D = a_Point2D.
        MathContext a_MathContext = new MathContext(_DecimalPlacePrecision + 5);
        BigDecimal xDiff0 = _End_Point2D._x.subtract(
                _Start_Point2D._x,
                a_MathContext);
        BigDecimal yDiff0 = _End_Point2D._y.subtract(
                _Start_Point2D._y,
                a_MathContext);
        BigDecimal xDiff1 = a_Point2D._x.subtract(
                _Start_Point2D._x,
                a_MathContext);
        BigDecimal yDiff1 = a_Point2D._y.subtract(
                _Start_Point2D._y,
                a_MathContext);
        BigDecimal gradient0;
        if (yDiff0.compareTo(BigDecimal.ZERO) == 0) {
            if (yDiff1.compareTo(BigDecimal.ZERO) == 0) {
                return true;
            } else {
                return false;
            }
        } else {
            gradient0 = xDiff0.divide(
                    yDiff0,
                    a_MathContext);
        }
        BigDecimal gradient1;
        if (yDiff1.compareTo(BigDecimal.ZERO) == 0) {
            return false;
        } else {
            gradient1 = xDiff1.divide(
                    yDiff1,
                    a_MathContext);
        }
        if (gradient0.compareTo(gradient1) == 0) {
            return true;
        }
        return false;
    }

    public LineSegment2D getOrderedLineSegment2D() {
        if (_Start_Point2D._y.compareTo(_End_Point2D._y) == -1) {
            return new LineSegment2D(this);
        } else {
            if (_Start_Point2D._y.compareTo(_End_Point2D._y) == 0) {
                if (_Start_Point2D._x.compareTo(_End_Point2D._x) != 1) {
                    return new LineSegment2D(this);
                }
            }
        }
        return new LineSegment2D(
                this._End_Point2D,
                this._Start_Point2D,
                this._DecimalPlacePrecision,
                this._RoundingMode);
    }

    /**
     * @return null if this does not intersect a_LineSegment2D; a Point2D if
     * this intersects a_LineSegment2D at a point; and, a LineSegment2D if this
     * intersects a_LineSegment2D in a line.
     */
    public AbstractGeometry2D getIntersection(
            LineSegment2D a_LineSegment2D) {
        MathContext a_MathContext = new MathContext(this._DecimalPlacePrecision + 100);
        BigDecimal x2minusx1 = _End_Point2D._x.subtract(_Start_Point2D._x, a_MathContext);
        BigDecimal y2minusy1 = _End_Point2D._y.subtract(_Start_Point2D._y, a_MathContext);
        BigDecimal x4minusx3 = a_LineSegment2D._End_Point2D._x.subtract(a_LineSegment2D._Start_Point2D._x, a_MathContext);
        BigDecimal y4minusy3 = a_LineSegment2D._End_Point2D._y.subtract(a_LineSegment2D._Start_Point2D._y, a_MathContext);
        BigDecimal denominator = (y4minusy3.multiply(x2minusx1, a_MathContext)).subtract(x4minusx3.multiply(y2minusy1, a_MathContext));
        boolean parallel = false;
        if (denominator.setScale(_DecimalPlacePrecision).compareTo(BigDecimal.ZERO) == 0) {
            System.out.println("parallel lines");
            parallel = true;
        }
        BigDecimal y1minusy3 = _Start_Point2D._y.subtract(a_LineSegment2D._Start_Point2D._y, a_MathContext);
        BigDecimal x1minusx3 = _Start_Point2D._x.subtract(a_LineSegment2D._Start_Point2D._x, a_MathContext);
        BigDecimal uamultiplicand = (x4minusx3.multiply(y1minusy3, a_MathContext)).subtract(y4minusy3.multiply(x1minusx3, a_MathContext), a_MathContext);
        BigDecimal ubmultiplicand = (x2minusx1.multiply(y1minusy3, a_MathContext)).subtract(y2minusy1.multiply(x1minusx3, a_MathContext), a_MathContext);
        if (uamultiplicand.setScale(_DecimalPlacePrecision).compareTo(BigDecimal.ZERO) == 0 &&
                ubmultiplicand.setScale(_DecimalPlacePrecision).compareTo(BigDecimal.ZERO) == 0 &&
                parallel) {
            System.out.println("lines coincident");
            LineSegment2D ot = getOrderedLineSegment2D();
            LineSegment2D oa = a_LineSegment2D.getOrderedLineSegment2D();
            boolean ts = ot._Start_Point2D.getIntersects(oa);
            boolean te = ot._End_Point2D.getIntersects(oa);
            boolean as = oa._Start_Point2D.getIntersects(ot);
            boolean ae = oa._End_Point2D.getIntersects(ot);
            if (ts) {
                if (te) {
                    if (as) {
                        if (ae) {
                            return new LineSegment2D(a_LineSegment2D);
                        } else {
                            return new LineSegment2D(
                                    ot._Start_Point2D,
                                    ot._End_Point2D);
                        }
                    } else {
                        if (ae) {
                            return new LineSegment2D(
                                    ot._Start_Point2D,
                                    oa._End_Point2D);
                        } else {
                            return new LineSegment2D(
                                    this);
                        }
                    }
                } else {
                    if (as) {
                        if (ae) {
                            return new LineSegment2D(a_LineSegment2D);
                        } else {
                            return new Point2D(ot._Start_Point2D);
                        }
                    } else {
                        System.out.println("Wierd...");
                        return null;
                    }
                }
            } else {
                if (te) {
                    if (as) {
                        if (ae) {
                            return new LineSegment2D(
                                    oa._Start_Point2D,
                                    oa._End_Point2D);
                        } else {
                            return new LineSegment2D(
                                    oa._Start_Point2D,
                                    ot._End_Point2D);
                        }
                    } else {
                        System.out.println("Wierd...");
                        return null;
                    }
                } else {
                    return new LineSegment2D(a_LineSegment2D);
                }
            }
        }
        if (parallel) {
            return null;
        }
        BigDecimal ua = uamultiplicand.divide(denominator, a_MathContext);
        if (ua.compareTo(BigDecimal.ONE) != 1 &&
                ua.compareTo(BigDecimal.ZERO) != -1) {
            BigDecimal intersectx1 = _Start_Point2D._x.add(ua.multiply(x2minusx1, a_MathContext), a_MathContext);
            BigDecimal intersecty1 = _Start_Point2D._y.add(ua.multiply(y2minusy1, a_MathContext), a_MathContext);
            BigDecimal ub = ubmultiplicand.divide(denominator, a_MathContext);
            return new Point2D(
                    intersectx1,
                    intersecty1,
                    _DecimalPlacePrecision + 10);
        } else {
            BigDecimal ub = ubmultiplicand.divide(denominator, a_MathContext);
            if (ub.compareTo(BigDecimal.ONE) != 1 &&
                    ub.compareTo(BigDecimal.ZERO) != -1) {
                BigDecimal intersectx2 = a_LineSegment2D._Start_Point2D._x.add(ub.multiply(x4minusx3, a_MathContext), a_MathContext);
                BigDecimal intersecty2 = a_LineSegment2D._Start_Point2D._y.add(ub.multiply(y4minusy3, a_MathContext), a_MathContext);
                return new Point2D(
                        intersectx2,
                        intersecty2,
                        _DecimalPlacePrecision + 10);
            }
        }
        return null;
    }

    @Override
    public void applyDecimalPlacePrecision() {
        this._Start_Point2D.applyDecimalPlacePrecision();
        this._End_Point2D.applyDecimalPlacePrecision();
    }
}
