package uk.ac.leeds.ccg.andyt.vector.geometry;

import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import uk.ac.leeds.ccg.andyt.grids.core.AbstractGrid2DSquareCell;

public class Network2D {

    public static double DefaultValue = 1;
    public static int DefaultType = 0;
    private static boolean HandleOutOfMemoryError = false;
    public HashMap<Point2D, HashSet<Connection>> _Connection_HashMap;
    //HashMap _Connection_HashMap;

    public Network2D() {
        _Connection_HashMap = new HashMap<Point2D, HashSet<Connection>>();
    }

    public Network2D(
            AbstractGrid2DSquareCell aGrid2DSquareCell) {
        _Connection_HashMap = new HashMap<Point2D, HashSet<Connection>>();
        Point2D a_Point2D;
        HashSet<Point2D> neighbouringPoints;
        for (long row = 0; row < aGrid2DSquareCell.get_NRows(HandleOutOfMemoryError); row++) {
            for (long col = 0; col < aGrid2DSquareCell.get_NCols(HandleOutOfMemoryError); col++) {
                a_Point2D = new Point2D(
                        aGrid2DSquareCell.getCellXBigDecimal(col, HandleOutOfMemoryError),
                        aGrid2DSquareCell.getCellYBigDecimal(row, HandleOutOfMemoryError));
                HashSet a_Connection_HashSet = new HashSet<Connection>();
                neighbouringPoints = getNeighbouringPoints(
                        aGrid2DSquareCell,
                        row,
                        col);
                Iterator a_Iterator = neighbouringPoints.iterator();
                while (a_Iterator.hasNext()) {
                    Point2D b_Point2D = (Point2D) a_Iterator.next();
                    Connection a_Connection = new Connection(
                            b_Point2D,
                            DefaultType,
                            DefaultValue);
                    a_Connection_HashSet.add(a_Connection);
                }
                _Connection_HashMap.put(
                        a_Point2D,
                        a_Connection_HashSet);
            }
        }
    }

    public void addToNetwork(
            Point2D toConnect_Point2D,
            HashSet<Connection> a_Connections_HashSet) {
        if (_Connection_HashMap.containsKey(toConnect_Point2D)) {
            ((HashSet) _Connection_HashMap.get(toConnect_Point2D)).addAll(a_Connections_HashSet);
        } else {
            _Connection_HashMap.put(
                    toConnect_Point2D,
                    a_Connections_HashSet);
        }
    }

    public void addToNetwork(
            Point2D toConnect_Point2D,
            Connection a_Connection) {
        if (_Connection_HashMap.containsKey(toConnect_Point2D)) {
            ((HashSet) _Connection_HashMap.get(toConnect_Point2D)).add(a_Connection);
        } else {
            HashSet<Connection> a_Connection_HashSet = new HashSet<Connection>();
            a_Connection_HashSet.add(a_Connection);
            _Connection_HashMap.put(
                    toConnect_Point2D,
                    a_Connection_HashSet);
        }
    }

    public void addToNetwork(
            Point2D toConnect_Point2D,
            Point2D toConnectTo_Point2D) {
        Connection a_Connection = new Connection(
                toConnectTo_Point2D,
                DefaultType,
                DefaultValue);
        addToNetwork(
                toConnect_Point2D,
                a_Connection);
    }

    @Override
    public String toString() {
        return "_Connection_HashMap.keySet().size() " + _Connection_HashMap.keySet().size();
    }

    public HashSet<Point2D> getNeighbouringPoints(
            AbstractGrid2DSquareCell aGrid2DSquareCell,
            long row,
            long col) {
        boolean _HandleOutOfMemoryError = false;
        HashSet<Point2D> result = new HashSet<Point2D>();
        long arow;
        long acol;
        arow = row - 1;
        acol = col - 1;
        if (aGrid2DSquareCell.isInGrid(arow, acol, _HandleOutOfMemoryError)) {
            result.add(
                    new Point2D(
                    aGrid2DSquareCell.getCellXBigDecimal(acol, _HandleOutOfMemoryError),
                    aGrid2DSquareCell.getCellYBigDecimal(arow, _HandleOutOfMemoryError)));

        }
        arow = row - 1;
        acol = col;
        if (aGrid2DSquareCell.isInGrid(arow, acol, _HandleOutOfMemoryError)) {
            result.add(
                    new Point2D(
                    aGrid2DSquareCell.getCellXBigDecimal(acol, _HandleOutOfMemoryError),
                    aGrid2DSquareCell.getCellYBigDecimal(arow, _HandleOutOfMemoryError)));

        }
        arow = row - 1;
        acol = col + 1;
        if (aGrid2DSquareCell.isInGrid(arow, acol, _HandleOutOfMemoryError)) {
            result.add(
                    new Point2D(
                    aGrid2DSquareCell.getCellXBigDecimal(acol, _HandleOutOfMemoryError),
                    aGrid2DSquareCell.getCellYBigDecimal(arow, _HandleOutOfMemoryError)));

        }
        arow = row;
        acol = col - 1;
        if (aGrid2DSquareCell.isInGrid(arow, acol, _HandleOutOfMemoryError)) {
            result.add(
                    new Point2D(
                    aGrid2DSquareCell.getCellXBigDecimal(acol, _HandleOutOfMemoryError),
                    aGrid2DSquareCell.getCellYBigDecimal(arow, _HandleOutOfMemoryError)));

        }
        arow = row;
        acol = col + 1;
        if (aGrid2DSquareCell.isInGrid(arow, acol, _HandleOutOfMemoryError)) {
            result.add(
                    new Point2D(
                    aGrid2DSquareCell.getCellXBigDecimal(acol, _HandleOutOfMemoryError),
                    aGrid2DSquareCell.getCellYBigDecimal(arow, _HandleOutOfMemoryError)));

        }
        arow = row + 1;
        acol = col - 1;
        if (aGrid2DSquareCell.isInGrid(arow, acol, _HandleOutOfMemoryError)) {
            result.add(
                    new Point2D(
                    aGrid2DSquareCell.getCellXBigDecimal(acol, _HandleOutOfMemoryError),
                    aGrid2DSquareCell.getCellYBigDecimal(arow, _HandleOutOfMemoryError)));

        }
        arow = row + 1;
        acol = col;
        if (aGrid2DSquareCell.isInGrid(arow, acol, _HandleOutOfMemoryError)) {
            result.add(
                    new Point2D(
                    aGrid2DSquareCell.getCellXBigDecimal(acol, _HandleOutOfMemoryError),
                    aGrid2DSquareCell.getCellYBigDecimal(arow, _HandleOutOfMemoryError)));

        }
        arow = row + 1;
        acol = col + 1;
        if (aGrid2DSquareCell.isInGrid(arow, acol, _HandleOutOfMemoryError)) {
            result.add(
                    new Point2D(
                    aGrid2DSquareCell.getCellXBigDecimal(acol, _HandleOutOfMemoryError),
                    aGrid2DSquareCell.getCellYBigDecimal(arow, _HandleOutOfMemoryError)));

        }
        return result;
    }

    public class Connection {

        public Point2D _Point2D;
        /**
         * Intended for capacity or other condition, or store of information
         * (e.g. number of flows)...
         */
        public double _Value;
        /**
         * To distinguish between different types of connection e.g. On foot,
         * by car etc...
         */
        public int _Type;

        public Connection() {
        }

        public Connection(
                Point2D _Point2D,
                int _Type,
                double _Value) {
            this._Point2D = _Point2D;
            this._Type = _Type;
            this._Value = _Value;
        }

        public void set_Value(double _Value) {
            this._Value = _Value;
        }
    }
}
